/*
** VoiceOver - a speech serializer for flite
**
** mqtt2vox  - relay Mosquitto MQTT messages to VoiceOver daemon
** Copyright (C) 2017  Lew Pitcher
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program; if not, write to the Free Software Foundation, Inc.,
** 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**
** Lew Pitcher <lew.pitcher@digitalfreehold.ca>
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <mosquitto.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <syslog.h>
#include <stdarg.h>
#include <errno.h>

#include "voxconfig.h"

/* for daemon() call		*/
#define DAEMON_CWD_TO_ROOT	0
#define DAEMON_CWD_AS_IS	1
#define DAEMON_CWD		DAEMON_CWD_TO_ROOT

#define DAEMON_STDIO_TO_NULL	0
#define DAEMON_STDIO_AS_IS	1
#define DAEMON_STDIO		DAEMON_STDIO_TO_NULL

struct mqttUserdata {
  char  *role;
  int	socket_fd;
  struct sockaddr_un serveraddress;
};

static volatile int DontStop = 1;	/* set to 0 by signal handler */
static unsigned int ReportingLevel = 0;

static void	receiveSignal(int);
static int 	Process(char *, char *, char *, char *, char *);
static void	handleMessage(struct mosquitto *, void *, const struct mosquitto_message *);
static void	LogToSyslog(unsigned int, int, const char *, ...);
static void	PidFile(char *);

int main(int argc, char *argv[])
{
  int	rc		= EXIT_FAILURE,
	validArgList	= 1,
	runInForeground = 1;
  char	*client_id	= NULL,
	*host		= NULL,
	*topic		= NULL,
	*s_path		= NULL,
	*pidpath	= NULL,
	*role		= NULL;

  int	opt;
  extern char *optarg;

  /* set up defaults */
  client_id	= argc > 0 ? argv[0]: NULL;
  s_path	= strdup(SOCKETPATH);

  while ((opt = getopt(argc,argv,"h:t:s:r:p:vd")) != -1)
  {
    switch (opt)
    {
      case 'h':	/* HOST NAME */
	if (host) free(host);
	host = strdup(optarg);
	break;

      case 't':	/* TOPIC NAME */
	if (topic) free(topic);
	topic = strdup(optarg);
	break;

      case 's':	/* SOCKET PATH */
	if (s_path) free(s_path);
	if ((s_path = realpath(optarg,NULL)) == NULL)
	  validArgList = 0;	/* cant resolve socket into a full path - maybe doesn't exist? */
	break;

      case 'r':	/* ROLE */
	if (role) free(role);
	role = strdup(optarg);
	break;

      case 'p':	/* PIDFILE PATH */
	if (*optarg == '/')	/* must be a fully qualified path */
	{
	  if (pidpath) free(pidpath);
	  pidpath = strdup(optarg);
	}
	else validArgList = 0;
	break;

      case 'd':	/* DAEMON */
	runInForeground = 0;
	break;

      case 'v':	/* Verbosity */
	++ReportingLevel;
	break;

      default:
	fprintf(stderr,"Unsupported option -%c\n",opt);
	validArgList = 0;
	break;
    }
  }

  if ((validArgList) && (argc == optind) && (host) && (topic))
  {
    if (runInForeground == 1 || daemon(DAEMON_CWD,DAEMON_STDIO) == 0)
    {
      if ((pidpath))	PidFile(pidpath);

      /* set up the "stop processing" signal handlers */
      signal(SIGINT,receiveSignal);
      signal(SIGTERM,receiveSignal);
      signal(SIGQUIT,receiveSignal);

      openlog(argv[0],LOG_PERROR|LOG_PID,LOG_DAEMON);
      if (Process(client_id,host,topic,s_path,role))
	rc = EXIT_SUCCESS;
      closelog();

      if ((pidpath))	unlink(pidpath);
    }
    else fprintf(stderr,"%s: cant daemonize - aborting\n",argv[0]?argv[0]:"program_name");
      
    free(host);		/* discard host name */
    free(topic);	/* discard topic */
    free(s_path);	/* discard socket path */
    free(role);		/* discard role */
    free(pidpath);	/* discard pidfile path */
  }
  else  fprintf(stderr,
		"Usage: %s -h <hostname> -t <topic> [-s <socketpath>] [-r <role>] [-p <pidfilepath>] [-d] [-v...]\n",
		argv[0]?argv[0]:"program_name");

  return rc;
}

static int Process(char *client_id, char *host, char *topic, char *s_path, char *role)
{
  int	rc		= 0;		/* 0 = failure, 1 = success */
  struct mosquitto *mq	= NULL;		/* mqtt connection handle */
  struct mqttUserdata userdata;		/* mqtt userdata workarea */

  if ((userdata.socket_fd = socket(AF_UNIX,SOCK_DGRAM,0)) >= 0)
  {
    memset(&userdata.serveraddress,0,sizeof userdata.serveraddress);
    userdata.serveraddress.sun_family = AF_LOCAL;
    strcpy(userdata.serveraddress.sun_path,s_path);
    userdata.role = role;

    mosquitto_lib_init();
    if ((mq = mosquitto_new(client_id,true,(void *)&userdata)))
    {
      mosquitto_message_callback_set(mq,handleMessage);
      if(mosquitto_connect(mq, host, 1883, 60) == MOSQ_ERR_SUCCESS)
      {
	if (mosquitto_subscribe(mq,NULL,topic,0) == MOSQ_ERR_SUCCESS)
	{
	  rc = 1;
	  while (DontStop) mosquitto_loop(mq, -1, 1);
          mosquitto_unsubscribe(mq,NULL,topic);
	}
 	else LogToSyslog(1,LOG_DAEMON|LOG_ERR,"Failed mosquitto_subscribe() to topic %s on host %s\n",topic,host);
	mosquitto_disconnect(mq);
      }
      else LogToSyslog(1,LOG_DAEMON|LOG_ERR,"Failed mosquitto_connect() to host %s\n",host);
      mosquitto_destroy(mq);
    }
    else LogToSyslog(1,LOG_DAEMON|LOG_ERR,"Failed mosquitto_new() with clientid %s\n",client_id?client_id:"[NULL]");
    mosquitto_lib_cleanup();
    close(userdata.socket_fd);
  }
  else LogToSyslog(1,LOG_DAEMON|LOG_ERR,"Failed socket() call: %s\n",strerror(errno));

  return rc;
}

/*
** Note: as this program uses the LAST component of the MQTT message's topic as
**	 the VoiceOver "ROLE" and the message payload as the VoiceOver "LINE", you
**	 can feed VoiceOver multiple lines for multiple roles by subscribing to
**	 a topic that ends in a wildcard.
**
**	 For example, a subscription to topic "VoiceOver/+" will assign the role "Butler"
**	 for messages addressed to topic "VoiceOver/Butler", and the role "Maid" for messages
**	 addressed to topic "VoiceOver/Maid"
*/
static void handleMessage(struct mosquitto *mq, void *mqttuserdata, const struct mosquitto_message *message)
{
  char	*topic = NULL,
	*subtopic = NULL,
	*role = NULL,
	*stringp = NULL,
	buffer[MAXMSGLEN+1];
  struct mqttUserdata *userdata;
  int	msgsize;

  userdata = mqttuserdata;

  if (userdata->role == NULL)
  {
    /* last component of message->topic is our Role */
    topic = strdup(message->topic);
    stringp = topic;
    while ((subtopic = strsep(&stringp,"/")))
      role = subtopic;
  }
  else role = userdata->role;

  LogToSyslog(2,LOG_DAEMON|LOG_INFO,"Received msg from %s, assigned role %s\n",message->topic,role);

  msgsize = snprintf(buffer,sizeof buffer,"%s:%*s",role,message->payloadlen,(char *)message->payload);
  if (topic) free(topic);

  if (msgsize <= MAXMSGLEN)
  {
    if (0 > sendto(userdata->socket_fd, buffer, msgsize, 0,
		   (struct sockaddr *)&userdata->serveraddress,sizeof userdata->serveraddress))
      LogToSyslog(1,LOG_DAEMON|LOG_ERR,"Failed sendto(): %s\n",strerror(errno));
  }
  else
    LogToSyslog(1,LOG_DAEMON|LOG_ERR,"%u byte msg from %s exceeds max length\n",msgsize, message->topic);
}

static void receiveSignal(int signum)
{
  signal(signum,SIG_IGN);
  DontStop = 0;				/* WE WANT TO STOP */
  signal(signum,receiveSignal);
}

static void LogToSyslog(unsigned int errorlevel, int priority, const char *format, ...)
{
  va_list	ap, 	/* our va_list */
		pp; 	/* vsyslog va_list */

  va_start(ap,format);
  if (errorlevel <= ReportingLevel)
  {
    va_copy(pp,ap);
    vsyslog(priority,format,pp);
  }
  va_end(ap);
}

static void PidFile(char *fpath)
{
  FILE *pidfile;

  pidfile = fopen(fpath,"w");
  fprintf(pidfile,"%lu\n",(unsigned long)getpid());
  fclose(pidfile);
}
