%{
  /* voxd.y - YACC grammar parser for voxd VoiceOver daemon config files
  ** Copyright (C) 2017  Lew Pitcher
  **
  ** This program is free software; you can redistribute it and/or modify
  ** it under the terms of the GNU General Public License as published by
  ** the Free Software Foundation; either version 2 of the License, or
  ** (at your option) any later version.
  **
  ** This program is distributed in the hope that it will be useful,
  ** but WITHOUT ANY WARRANTY; without even the implied warranty of
  ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  ** GNU General Public License for more details.
  **
  ** You should have received a copy of the GNU General Public License along
  ** with this program; if not, write to the Free Software Foundation, Inc.,
  ** 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  **
  ** Lew Pitcher <lew.pitcher@digitalfreehold.ca>
  */

  /* headers, globals, etc */
  #include <stdio.h>
  #include <stdlib.h>
  #include <string.h>
  #include <syslog.h>
  #include "voxd.h"

  unsigned int LineNumber = 1;

  void yyerror(char *);

  extern int yylex(void);


  static struct actor	*newActor(void);
  static struct actor	*findActor(char *);
  static struct role	*newRole(void);
  static struct role	*findRole(char *);

  static float	currTempo = 0.0,
		currPitch = 0.0,
		currIntonation = 0.0;

  static struct show  *Show = NULL;
  static struct actor *actorList = NULL;
  static struct role  *roleList = NULL;
  static char	*prologueTxt = NULL,
		*epilogueTxt = NULL;

%}
%token PROLOGUE EPILOGUE TEMPO PITCH INTONATION VOICEDBY HASVOICE LPAREN RPAREN EQUALS COMMA
%union
{
  struct actor	*Actor;
  char		*String;
  float		Float;
}
%token <String> QUOTEDSTRING NAME
%token <Float>  FLOATPT
%type  <Actor>  actor
%%
vox	: lines
	{
	  if (findRole("STANDIN") == NULL)
	  {
	    struct actor *a_standin;
	    struct role *r_standin;

	    if ((a_standin = newActor()))
	    {
	      if ((r_standin = newRole()))
	      {
		r_standin->roleName = strdup(STANDIN_ROLE);
		r_standin->roleActor = a_standin;
		a_standin->actorName = strdup(STANDIN_ACTOR);
		a_standin->voicePath = strdup(STANDIN_VOICE);

		r_standin->roleNext = roleList;		roleList = r_standin;
		a_standin->actorNext = actorList;	actorList = a_standin;
	      }
	      else
	      {
		free(a_standin);
		yyerror("cant allocate space for STANDIN role");
	      }
	    }
	    else yyerror("cant allocate space for STANDIN actor");
	  }

	  if ((Show = malloc(sizeof *Show)))
	  {
	    Show->roles  = roleList;	roleList  = NULL;
	    Show->actors = actorList;	actorList = NULL;
	    Show->prologue = prologueTxt;	prologueTxt  = NULL;
	    Show->epilogue = epilogueTxt;	epilogueTxt  = NULL; 
	  }
	  else yyerror("cant allocate space for show");
	}
	;

lines	: lines line
	| line
	| error '\n'
	;

line	: PROLOGUE EQUALS QUOTEDSTRING
	{
	  prologueTxt = $3;
	}
	| EPILOGUE EQUALS QUOTEDSTRING
	{
	  epilogueTxt = $3;
	}
	| NAME hasvoice actor
	{
	  if (findActor($1) == NULL)
	  {
	    $3->actorName = $1;
	    $3->actorNext = actorList;
	    actorList = $3;
	  }
	  else
	  {
	    free($3->voicePath);
	    free($3);
	    yyerror("duplicate actor name - actor discarded");
	  }
	}
	| NAME voicedby NAME
	{
	  struct role *thisrole;

	  if ((thisrole = findRole($1)) == NULL)
	  {
	    struct actor *thisactor;
	    if ((thisactor = findActor($3)))
	    {
	      if ((thisrole = newRole()))
	      {
	        thisrole->roleName = $1;
		thisrole->roleActor = thisactor;
		thisrole->roleNext = roleList;
		roleList = thisrole;
	      }
	      else yyerror("out of space while allocating role");
	    }
	    else yyerror("named actor not yet defined");
	  }
	  else yyerror("duplicate role name - role discarded");
	  
	}
	;

hasvoice : HASVOICE
	 | EQUALS
	 ;

voicedby : VOICEDBY
	 | EQUALS
	 ;

actor	: QUOTEDSTRING LPAREN mods RPAREN
	{
	  struct actor *actor;

	  if ((actor = newActor()))
	  {
	    actor->voicePath = $1;
	    actor->voiceTempo = currTempo;	currTempo = 0.0;
	    actor->voicePitch = currPitch;	currPitch = 0.0;
	    actor->voiceIntonation = currIntonation;	currIntonation = 0.0;
	  }
	  else yyerror("out of space while allocating actor");
	  $$ = actor;
	}
	| QUOTEDSTRING
	{
	  struct actor *actor;

	  if ((actor = newActor()))
	  {
	    actor->voicePath = $1;
	  }
	  else yyerror("out of space while allocating actor");
	  $$ = actor;
	}
	;

mods	: mods COMMA mod
	| mod
	;

mod	: TEMPO EQUALS FLOATPT
	{
	  if (currTempo == 0.0)
	    currTempo = $3;
	  else yyerror("excess tempo value ignored");
	}
	| PITCH EQUALS FLOATPT
	{
	  if (currPitch == 0.0)
	    currPitch = $3;
	  else yyerror("excess pitch value ignored");
	}
	| INTONATION EQUALS FLOATPT
	{
	  if (currIntonation == 0.0)
	    currIntonation = $3;
	  else yyerror("excess intonation value ignored");
	}
	;
%%
void yyerror(char *msg)
{
  extern char *yytext;

#ifdef LOG_TO_STDERR
  fprintf(stderr,"Line %u, token \"%s\": %s\n",LineNumber,yytext,msg);
#else
  syslog(LOG_ERR,"Line %u, token \"%s\": %s\n",LineNumber,yytext,msg);
#endif
}

int yywrap()
{
  return 1;
}

static struct actor *newActor(void)
{
  struct actor *new;

  if ((new = malloc(sizeof *new)))
  {
    new->actorNext = NULL;
    new->actorName = NULL;
    new->voicePath = NULL;
    new->voiceTempo = 0.0;
    new->voicePitch = 0.0;
    new->voiceIntonation = 0.0;
  }
  return new;
}

static struct actor *findActor(char *name)
{
  struct actor *this;

  for (this = actorList; this && strcmp(this->actorName,name) != 0; this = this->actorNext) continue;
  return this;
}

static struct role *newRole(void)
{
  struct role *new;

  if ((new = malloc(sizeof *new)))
  {
    new->roleNext = NULL;
    new->roleName = NULL;
    new->roleActor = NULL;
  }
  return new;
}

static struct role *findRole(char *name)
{
  struct role *this;

  for (this = roleList; this && strcmp(this->roleName,name) != 0; this = this->roleNext) continue;
  return this;
}

/*
********************************************************************************
*/

/*
** OpenShow()	Construct VoiceOver configuration tree
** Accepts:	pointer to character string naming configuration file path
** Process:	read config file to build config tree
** Returns:	pointer to the root of config tree
*/
struct show *OpenShow(char *configpath)
{
  extern FILE *yyin;
  struct show *temp = NULL;

  if ((yyin = fopen(configpath,"r")))
  {
    if (yyparse() == 0)
    {
      temp = Show;
      Show = NULL;
    }
    else fputs("Cannot parse config file\n",stderr);
    fclose(yyin);
  }
  else perror(configpath);

  return temp;
}

/*
** CloseShow()	Demolish VoiceOver configuration tree
** Accepts:	pointer to the root of config tree
** Process:	release all storage allocated to the tree
** Returns:	nothing
*/
void CloseShow(struct show *Show)
{
  struct role  *Role;
  struct actor *Actor;

  for (Role = Show->roles; Role;)
  {
    struct role *next;

    next = Role->roleNext;
    free(Role->roleName);
    free(Role);
    Role = next;
  }

  for (Actor = Show->actors; Actor;)
  {
    struct actor *next;

    next = Actor->actorNext;
    free(Actor->actorName);
    free(Actor->voicePath);
    free(Actor);
    Actor = next;
  }

  free(Show->prologue);
  free(Show->epilogue);
  free(Show);
}

/*
** GetActorInRole	Retrieve voice particulars for a specified role
** Accepts:	pointer to character string naming the role to retrieve
**		pointer to struct show of root of config tree
** Process:	look through config tree for role
**		if found, return the actor for that role
**		otherwise, look through config tree for actor
**		if found, return the actor /named/ by the role
**		othewise, return the actor that voices for STANDIN
** Returns:	pointer to struct actor of the voice for the specified role
*/
struct actor *GetActorInRole(char *Role, struct show *Show)
{
  struct actor	*actor = NULL,
		*standin = NULL;
  struct role	*role;

  for (role = Show->roles; role; role = role->roleNext)
  {
    if (strcmp(STANDIN_ROLE,role->roleName) == 0)	standin = role->roleActor;
    if (strcmp(Role,role->roleName) == 0)	actor = role->roleActor;
  }

  if (actor == NULL)	/* No actor for this role, perhaps role /is/ actor? */
    for (actor = Show->actors; actor && (strcmp(actor->actorName,Role) != 0); actor = actor->actorNext)
      continue;	/* search list of actors for an actor who's name matches the requested Role */

  if (actor == NULL)	/* couldn't find the proper actor */ 
    actor = standin;	/* so use the standin instead	  */

  return actor;
}

