/*
** VoiceOver - a speech serializer for flite
**
** voxd_main	mainline for voxd VoiceOver daemon
** Copyright (C) 2017  Lew Pitcher
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program; if not, write to the Free Software Foundation, Inc.,
** 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**
** Lew Pitcher <lew.pitcher@digitalfreehold.ca>
*/

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <syslog.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/select.h>

#include "voxconfig.h"
#include "voxd.h"
#include "flite.h"

/* for daemon() call		*/
#define DAEMON_CWD_TO_ROOT	0
#define DAEMON_CWD_AS_IS	1
#define DAEMON_CWD		DAEMON_CWD_TO_ROOT

#define DAEMON_STDIO_TO_NULL	0
#define DAEMON_STDIO_AS_IS	1
#define DAEMON_STDIO		DAEMON_STDIO_TO_NULL


/* for select() call		*/
#define	WAIT_SECS		15
#define	WAIT_USEC		0


/* PAUSE this long between role changes */
#define	PAUSE			2

static volatile int Processing = 1;		/* normally 1, set to 0 when termination requested */
static volatile int ReloadConfig  = 0;		/* normally 0, set to 1 when config file reload requested */

static void RequestStop(int);			/* interrupt handler - clears Processing flag */
static void RefreshConfig(int);			/* interrupt handler - sets ReloadConfig flag */
static int  GetArgs(int , char **, char **, int *, char **, char **, char **);
static void PidFile(char *);			/* record pid in file - for external signalling */

static int VoiceOver(char *, char *);		/* main processing routine - run in daemon */
static void ReadLine(struct show *, char *, char *);	/* speaks text out loud */

int main(int argc, char *argv[])
{
  int	rc = EXIT_FAILURE;
  char	*ConfigPath	= NULL,
	*LockFilePath	= NULL,
	*PidFilePath	= NULL,
	*SocketPath	= NULL;
  int	runInForeground;

  if (GetArgs(argc,argv,&ConfigPath,&runInForeground,&LockFilePath,&PidFilePath,&SocketPath) == 0)
  {
    int lockfile;
    openlog(argv[0],LOG_PID,LOG_DAEMON);

    if ((lockfile = open(LockFilePath,O_WRONLY|O_CREAT|O_EXCL,0600)) >= 0)
    {
      syslog(LOG_INFO,"Acquiring shared lock on %s\n",LockFilePath);

      if (flock(lockfile,LOCK_SH|LOCK_NB) == 0)
      {
	syslog(LOG_INFO,"Daemonizing\n");

	if (runInForeground == 1 || daemon(DAEMON_CWD,DAEMON_STDIO) == 0)
	{
	  syslog(LOG_INFO,"Acquiring exclusive lock on %s\n",LockFilePath);

	  if (flock(lockfile,LOCK_EX|LOCK_NB) == 0)
	  {
	    syslog(LOG_INFO,"Processing starts\n");

	    PidFile(PidFilePath);
	    if (VoiceOver(ConfigPath,SocketPath)) rc = EXIT_SUCCESS;
	    unlink(PidFilePath);

	    syslog(LOG_INFO,"Processing complete\n");
	  }
	  else syslog(LOG_CRIT,"Cant acquire exclusive lock on %s - aborting\n",LockFilePath);
	}
	else syslog(LOG_CRIT,"Cant daemonize - aborting\n");

	syslog(LOG_INFO,"Discarding lock file %s and releasing lock\n",LockFilePath);
	unlink(LockFilePath);
	flock(lockfile,LOCK_UN);
	close(lockfile);
      }
      else  syslog(LOG_CRIT,"Cant acquire shared lock on %s - aborting\n",LockFilePath);
    }
    else syslog(LOG_CRIT,"Cant allocate lockfile %s - aborting\n",LockFilePath);

    closelog();
  }
  else syslog(LOG_CRIT,"Invalid commandline options- aborting\n");

  return rc;
}

/*
** configuration that can be altered from the commandline:
** - the path to the config file
**	-c <configpath>
** - whether or not the program will daemonize
**	-b
** - the path to the lockfile
**	-l <lockfilepath>
** - the path to the pidfile
**	-p <pidfilepath>
** - the path to the unix socket
**	-s <socketfilepath>
*/
static int GetArgs(int argc, char *argv[],
	    char **configpath,
	    int *foreground,
	    char **lockfilepath,
	    char **pidfilepath,
	    char **socketpath)
{
  int	opt;
  extern char *optarg;

  char	*ConfigPath	= NULL,
	*LockFilePath	= NULL,
	*PidFilePath	= NULL,
	*SocketPath	= NULL;
  int	runInForeground = 1,
	argerrors	= 0;

  ConfigPath	 = strdup(CONFIGPATH);
  LockFilePath	 = strdup(LOCKFILEPATH);
  PidFilePath	 = strdup(PIDFILEPATH);
  SocketPath	 = strdup(SOCKETPATH);

  while ((opt = getopt(argc,argv,"c:l:p:s:d")) != -1)
  {
    switch (opt)
    {
      case 'c':	/* change configfile path */
	free(ConfigPath);
	if ((ConfigPath = realpath(optarg,NULL)) == NULL)
	  argerrors = 1; /* cant resolve config into a full path - maybe doesn't exist? */
	break;

      case 'd':	/* run as daemon */
	runInForeground = 0;
	break;

      case 'l':	/* change lockfile path */
	if (*optarg == '/')	/* must be a fully qualified path */
	{
	  free(LockFilePath);
	  LockFilePath = strdup(optarg);
	}
	else argerrors = 1;
	break;

      case 'p':	/* change pidfile path */
	if (*optarg == '/')	/* must be a fully qualified path */
	{
	  free(PidFilePath);
	  PidFilePath = strdup(optarg);
	}
	else argerrors = 1;
	break;

      case 's':	/* change socket path */
	if (*optarg == '/')	/* must be a fully qualified path */
	{
	  free(SocketPath);
	  SocketPath = strdup(optarg);
	}
	else argerrors = 1;
	break;

      default:	/* anything else */
	argerrors = 1;
	break;
    }
  }

  if (argerrors == 0)
  {
    *configpath   = ConfigPath;
    *foreground   = runInForeground;
    *lockfilepath = LockFilePath;
    *pidfilepath  = PidFilePath;
    *socketpath   = SocketPath;
  }
  else fprintf(stderr,"Usage: %s [-c configpath] [-d] [-l lockfilepath] [-p pidfilepath] [-s socketpath]\n",argv[0]);

  return argerrors;
}


static void PidFile(char *fpath)
{
  FILE *pidfile;

  pidfile = fopen(fpath,"w");
  fprintf(pidfile,"%lu\n",(unsigned long)getpid());
  fclose(pidfile);
}

static int VoiceOver(char *ConfigPath, char *Socket)
{
  int rc = 0;		/* default to Failed */
  struct show *Show;

  syslog(LOG_INFO,"Getting Show from config %s\n",ConfigPath);

  if ((Show = OpenShow(ConfigPath)))
  {
    int script;
    mode_t old_umask; 

    old_umask = umask(0155);

    syslog(LOG_INFO,"Establishing UNIX datagram socket\n");
    if ((script = socket(AF_UNIX,SOCK_DGRAM,0)) >= 0)
    {
      struct sockaddr_un serveraddress;

      syslog(LOG_INFO,"Binding UNIX datagram socket to %s\n",Socket);
      unlink(Socket);
      memset(&serveraddress,0,sizeof serveraddress);
      serveraddress.sun_family = AF_LOCAL;
      strcpy(serveraddress.sun_path,Socket);
      if (bind(script,(struct sockaddr *)&serveraddress,sizeof serveraddress) == 0)
      {
	char	*prevRole = NULL;
	time_t	timePrev = 0,
		timeDiff;

	rc = 1;		/* Succeded */

	InitFlite();

	/* establish signal handlers to set the refresh-config and stop-looping flags */
	signal(SIGHUP, RefreshConfig);
	signal(SIGINT, RequestStop);
	signal(SIGTERM,RequestStop);

	if (Show->prologue) ReadLine(Show,"",Show->prologue);
	while (Processing)
	{
	  fd_set rfds;
	  struct timeval tv;

	  char	buffer[MAXMSGLEN+1],
		*Line;
	  ssize_t msglen;

	  if (ReloadConfig)
	  {
	    struct show *NewShow;

	    if ((NewShow = OpenShow(ConfigPath)))
	    {
	      syslog(LOG_INFO,"Got new Show from %s\n",ConfigPath);
	      CloseShow(Show);
	      Show = NewShow;
	    }
	    ReloadConfig = 0;
	  }

	  /* wait until line ready or timer goes off */
	  FD_ZERO(&rfds); FD_SET(script,&rfds);
	  tv.tv_sec = WAIT_SECS; tv.tv_usec = WAIT_USEC;
	  if (select(script+1,&rfds,NULL,NULL,&tv) > 0)
	  {
	    timeDiff = time(NULL) - timePrev; 

	    msglen = read(script,buffer,(sizeof buffer)-1);
	    buffer[msglen] = 0;
	    if (msglen > 0 && (Line = strchr(buffer,':')) != NULL)
	    {
	      *Line++  = 0;	/* make first part of buffer into role name string */
	      while (isspace(*Line)) ++Line;	/* skip leading whitespace in Line */

	      /* When the role changes, we insert a pause to reduce listener confusion */
	      if (prevRole && strcmp(prevRole,buffer) != 0 && timeDiff < PAUSE) sleep(PAUSE - timeDiff);
	      free(prevRole); prevRole = strdup(buffer);

	      ReadLine(Show,buffer,Line);
	      timePrev = time(NULL);
	    }
	  }
	}
	if (Show->epilogue) ReadLine(Show,"",Show->epilogue);
	close(script);
	unlink(Socket);
      }
      else  syslog(LOG_CRIT,"Failed to bind socket to %s - aborting\n",Socket);
    }
    else syslog(LOG_CRIT,"Failed to create UNIX DATAGRAM socket - aborting\n");
    umask(old_umask);
  }
  else syslog(LOG_CRIT,"Failed to obtain configuration from %s - aborting\n",ConfigPath);
  return rc;
}

static void ReadLine(struct show *Cast, char *Role, char *Line)
{
  struct actor *Actor;

  if ((Actor = GetActorInRole(Role,Cast)))
  {
    cst_voice *Voice;

    if ((Voice = GetVoice(Actor->voicePath,Actor->voiceTempo,Actor->voicePitch,Actor->voiceIntonation)))
      Say(Voice,"%s",Line);
    else
      syslog(LOG_ERR,"Actor %s has no voice and cant say the line\n",Actor->actorName);
  }
  else syslog(LOG_ERR,"Role %s has no actor and the line remains unsaid\n",Role);
}

static void RequestStop(int signum)
{
  signal(signum,SIG_IGN);
  syslog(LOG_INFO,"Received request to shut down\n");
  Processing = 0;			/* WE WANT TO STOP */
  signal(signum,RequestStop);
}

static void RefreshConfig(int signum)
{
  signal(signum,SIG_IGN);
  syslog(LOG_INFO,"Received request to load new configuration\n");
  ReloadConfig = 1;
  signal(signum,RefreshConfig);
}
